package aligner;

import java.io.File;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import dna.AminoAcid;
import fileIO.ByteFile;
import fileIO.ByteStreamWriter;
import fileIO.FileFormat;
import fileIO.ReadWrite;
import shared.Parse;
import shared.Parser;
import shared.PreParser;
import shared.SIMDAlignByte;
import shared.Shared;
import shared.Timer;
import shared.Tools;
import stream.ConcurrentReadInputStream;
import stream.FastaReadInputStream;
import stream.Read;
import stream.SamLine;
import structures.ByteBuilder;
import structures.IntHashMap;
import structures.IntList;
import structures.IntListHashMap;
import structures.ListNum;
import template.Accumulator;
import template.ThreadWaiter;
import tracker.ReadStats;

/**
 * Performs indel-free alignments of queries.
 * Designed for a small query set with stays in memory,
 * and a large reference set which is streamed.
 * Allows an arbitrary number of substitutions.
 * Uses SIMD.
 * 
 * @author Brian Bushnell
 * @contributor Isla
 * @date June 2, 2025
 *
 */
public class IndelFreeAligner implements Accumulator<IndelFreeAligner.ProcessThread> {
	
	/*--------------------------------------------------------------*/
	/*----------------        Initialization        ----------------*/
	/*--------------------------------------------------------------*/
	
	/**
	 * Code entrance from the command line.
	 * @param args Command line arguments
	 */
	public static void main(String[] args){
		//Start a timer immediately upon code entrance.
		Timer t=new Timer();
		
		//Create an instance of this class
		IndelFreeAligner x=new IndelFreeAligner(args);
		
		//Run the object
		x.process(t);
		
		//Close the print stream if it was redirected
		Shared.closeStream(x.outstream);
	}
	
	/**
	 * Constructor.
	 * @param args Command line arguments
	 */
	public IndelFreeAligner(String[] args){
		
		{//Preparse block for help, config files, and outstream
			PreParser pp=new PreParser(args, getClass(), false);
			args=pp.args;
			outstream=pp.outstream;
		}
		
		//Set shared static variables prior to parsing
		ReadWrite.USE_PIGZ=ReadWrite.USE_UNPIGZ=true;
		ReadWrite.setZipThreads(Shared.threads());
		
		{//Parse the arguments
			final Parser parser=parse(args);
			Parser.processQuality();
			
			maxReads=parser.maxReads;
			overwrite=ReadStats.overwrite=parser.overwrite;
			append=ReadStats.append=parser.append;
			
			in1=parser.in1;
			in2=parser.in2;
			extin=parser.extin;

			out1=parser.out1;
			extout=parser.extout;
		}

		validateParams();
		doPoundReplacement(); //Replace # with 1 and 2
		fixExtensions(); //Add or remove .gz or .bz2 as needed
		checkFileExistence(); //Ensure files can be read and written
		checkStatics(); //Adjust file-related static fields as needed for this program 
		
		//Create output FileFormat objects
		ffout1=FileFormat.testOutput(out1, FileFormat.FASTQ, extout, true, overwrite, append, false);

		//Create input FileFormat objects
		ffin1=FileFormat.testInput(in1, FileFormat.FASTQ, extin, true, true);
		ffin2=FileFormat.testInput(in2, FileFormat.FASTQ, extin, true, true);
	}
	
	/*--------------------------------------------------------------*/
	/*----------------    Initialization Helpers    ----------------*/
	/*--------------------------------------------------------------*/
	
	/** Parse arguments from the command line */
	private Parser parse(String[] args){
		
		//Create a parser object
		Parser parser=new Parser();
		
		//Set any necessary Parser defaults here
		//parser.foo=bar;
		
		//Parse each argument
		for(int i=0; i<args.length; i++){
			String arg=args[i];
			
			//Break arguments into their constituent parts, in the form of "a=b"
			String[] split=arg.split("=");
			String a=split[0].toLowerCase();
			String b=split.length>1 ? split[1] : null;
			if(b!=null && b.equalsIgnoreCase("null")){b=null;}
			
			if(a.equals("verbose")){
				verbose=Parse.parseBoolean(b);
			}else if(a.equals("ref")){
				refFile=b;
			}else if(a.equals("subs") || a.equals("maxsubs")){
				maxSubs=Integer.parseInt(b);
			}else if(a.equals("hits") || a.equals("minhits") || a.equals("seedhits")){
				minSeedHits=Math.max(1, Integer.parseInt(b));
			}else if(a.equals("minprob") || a.equals("minhitsprob")){
				minHitsProb=Float.parseFloat(b);
			}else if(a.equals("index")){
				indexQueries=Parse.parseBoolean(b);
			}else if(a.equals("seedmap") || a.equals("map")){
				useSeedMap=Parse.parseBoolean(b);
			}else if(a.equals("seedlist") || a.equals("list")){
				useSeedMap=!Parse.parseBoolean(b);
			}else if(a.equals("k")){
				k=Integer.parseInt(b);
				assert(k<16) : "0<=k<16 : "+k;
				indexQueries=(k>0);
			}else if(a.equals("qstep") || a.equals("step") || a.equals("qskip")){
				qStep=Integer.parseInt(b);
				assert(qStep>0);
			}else if(a.equals("mm")){
				midMaskLen=(Tools.isNumeric(b) ? Integer.parseInt(b) :
					Parse.parseBoolean(b) ? 1 : 0);
			}else if(a.equals("blacklist") || a.equals("banhomopolymers")){
				Query.blacklistRepeatLength=(Tools.isNumeric(b) ? Integer.parseInt(b) :
					Parse.parseBoolean(b) ? 1 : 0);
			}else if(a.equals("parse_flag_goes_here")){
				long fake_variable=Parse.parseKMG(b);
				//Set a variable here
			}else if(parser.parse(arg, a, b)){//Parse standard flags in the parser
				//do nothing
			}else{
				outstream.println("Unknown parameter "+args[i]);
				assert(false) : "Unknown parameter "+args[i];
			}
		}
		
		return parser;
	}
	
	/** Replace # with 1 and 2 in headers */
	private void doPoundReplacement(){
		//Do input file # replacement
		if(in1!=null && in2==null && in1.indexOf('#')>-1 && !new File(in1).exists()){
			in2=in1.replace("#", "2");
			in1=in1.replace("#", "1");
		}
		
		//Ensure there is an input file
		if(in1==null){throw new RuntimeException("Error - at least one input file is required.");}
	}
	
	/** Add or remove .gz or .bz2 as needed */
	private void fixExtensions(){
		in1=Tools.fixExtension(in1);
		in2=Tools.fixExtension(in2);
	}
	
	/** Ensure files can be read and written */
	private void checkFileExistence(){
		//Ensure output files can be written
		if(!Tools.testOutputFiles(overwrite, append, false, out1)){
			outstream.println((out1==null)+", "+out1);
			throw new RuntimeException("\n\noverwrite="+overwrite+"; Can't write to output files "+out1+"\n");
		}
		
		//Ensure input files can be read
		if(!Tools.testInputFiles(false, true, in1, in2)){
			throw new RuntimeException("\nCan't read some input files.\n");  
		}
		
		//Ensure that no file was specified multiple times
		if(!Tools.testForDuplicateFiles(true, in1, in2, out1)){
			throw new RuntimeException("\nSome file names were specified multiple times.\n");
		}
	}
	
	/** Adjust file-related static fields as needed for this program */
	private static void checkStatics(){
		//Adjust the number of threads for input file reading
		if(!ByteFile.FORCE_MODE_BF1 && !ByteFile.FORCE_MODE_BF2 && Shared.threads()>2){
			ByteFile.FORCE_MODE_BF2=true;
		}
		
		assert(FastaReadInputStream.settingsOK());
	}
	
	/** Ensure parameter ranges are within bounds and required parameters are set */
	private boolean validateParams(){
	    assert((k>=1 && k<=15) || !indexQueries);
	    assert(minHitsProb<=1);
	    assert(midMaskLen<k-1);
	    assert(maxSubs>=0);
	    return ((k>=1 && k<=15) || !indexQueries) && 
	    		(minHitsProb<=1) && (midMaskLen<k-1) && (maxSubs>=0);
	}
	
	/*--------------------------------------------------------------*/
	/*----------------         Outer Methods        ----------------*/
	/*--------------------------------------------------------------*/

	/** Create read streams and process all data */
	void process(Timer t){
		
		//Reset counters
		readsProcessed=readsOut=0;
		basesProcessed=basesOut=0;

		SamLine.RNAME_AS_BYTES=false;
		
		Query.setMode(k, midMaskLen, indexQueries);
		final ArrayList<Query> queries=fetchQueries(ffin1, ffin2);
		
		//Turn off read validation in the input threads to increase speed
		final boolean vic=Read.VALIDATE_IN_CONSTRUCTOR;
		Read.VALIDATE_IN_CONSTRUCTOR=Shared.threads()<4;
		
		//Create a read input stream
		final ConcurrentReadInputStream cris=makeCris(refFile);
		
		//Optionally create a read output stream
		final ByteStreamWriter bsw=ByteStreamWriter.makeBSW(ffout1);
		
		//Process the reads in separate threads
		spawnThreads(cris, bsw, queries);
		
		if(verbose){outstream.println("Finished; closing streams.");}
		
		//Write anything that was accumulated by ReadStats
		errorState|=ReadStats.writeAll();
		//Close the read streams
		errorState|=ReadWrite.closeStreams(cris);
		
		if(bsw!=null) {bsw.poisonAndWait();}
		
		//Reset read validation
		Read.VALIDATE_IN_CONSTRUCTOR=vic;
		
		//Report timing and results
		t.stop();
		outstream.println(Tools.timeReadsBasesProcessed(t, readsProcessed, basesProcessed, 8));
		outstream.println(Tools.readsBasesOut(readsProcessed, basesProcessed, readsOut, basesOut, 8, false));
		outstream.println(Tools.things("Alignments", alignmentCount, 8));
		outstream.println(Tools.things("Seed Hits", seedHitCount, 8));
		
		//Throw an exception of there was an error in a thread
		if(errorState){
			throw new RuntimeException(getClass().getName()+" terminated in an error state; the output may be corrupt.");
		}
	}
	
	/*--------------------------------------------------------------*/
	/*----------------       Thread Management      ----------------*/
	/*--------------------------------------------------------------*/
	
	/** Spawn process threads */
	private void spawnThreads(final ConcurrentReadInputStream cris, 
			final ByteStreamWriter bsw, final ArrayList<Query> qList){
		
		//Do anything necessary prior to processing
		
		//Determine how many threads may be used
		final int threads=Shared.threads();
		
		//Fill a list with ProcessThreads
		ArrayList<ProcessThread> alpt=new ArrayList<ProcessThread>(threads);
		for(int i=0; i<threads; i++){
			alpt.add(new ProcessThread(cris, bsw, qList, maxSubs, i));
		}
		
		//Start the threads and wait for them to finish
		boolean success=ThreadWaiter.startAndWait(alpt, this);
		errorState&=!success;
		
		//Do anything necessary after processing
		
	}
	
	@Override
	public final void accumulate(ProcessThread pt){
		synchronized(pt) {
			readsProcessed+=pt.readsProcessedT;
			basesProcessed+=pt.basesProcessedT;
			alignmentCount+=pt.alignmentsT;
			seedHitCount+=pt.seedHitsT;
			
			readsOut+=pt.readsOutT;
			basesOut+=pt.basesOutT;
			errorState|=(!pt.success);
		}
	}
	
	@Override
	public final boolean success(){return !errorState;}
	
	/*--------------------------------------------------------------*/
	/*----------------         Inner Methods        ----------------*/
	/*--------------------------------------------------------------*/
	
	private ConcurrentReadInputStream makeCris(String fname){
		FileFormat ff=FileFormat.testInput(fname, null, true);
		ConcurrentReadInputStream cris=ConcurrentReadInputStream.getReadInputStream(maxReads, true, ff, null);
		cris.start(); //Start the stream
		if(verbose){outstream.println("Started cris");}
		return cris;
	}
	
	public ArrayList<Query> fetchQueries(FileFormat ff1, FileFormat ff2){
		Timer t=new Timer(outstream, false);
		ArrayList<Read> reads=ConcurrentReadInputStream.getReads(maxReads, false, ff1, ff2, null, null);
		ArrayList<Query> queries=new ArrayList<Query>(reads.size());
		if(indexQueries) {
			Query.mhc=new MinHitsCalculator(k, maxSubs, midMaskLen, minHitsProb);
		}
		for(Read r : reads) {//TODO: Could be multithreaded.
			readsProcessed+=r.pairCount();
			basesProcessed+=r.pairLength();
			Query query=new Query(r.id, queries.size(), r.bases, r.quality);
			queries.add(query);
			if(r.mate==null) {continue;}
			r=r.mate;
			query=new Query(r.id, queries.size(), r.bases, r.quality);
			queries.add(query);
		}
		t.stop("Loaded "+queries.size()+" queries in ");
		return queries;
	}
	
	//Opus version; 10% slower
//	public static IntList align(byte[] query, byte[] ref, int maxSubs) {
//	    IntList list=null;
//	    final int start = -maxSubs;
//	    final int end = ref.length - query.length + maxSubs;
//	    
//	    for(int rStart = start; rStart <= end; rStart++) {
//	        int subs = alignSimple(query, ref, maxSubs, rStart);
//	        if(subs <= maxSubs) {
//	            if(list == null) {list = new IntList(4);}
//	            list.add(rStart);
//	        }
//	    }
//	    return list;
//	}
//
//	private static int alignSimple(byte[] query, byte[] ref, int maxSubs, final int rStart) {
//	    int subs = 0;
//	    for(int i = 0, j = rStart; i < query.length && subs <= maxSubs; i++, j++) {
//	        final byte q = query[i];
//	        final byte r = (j >= 0 && j < ref.length) ? ref[j] : (byte)'N';
//	        if(q != r || !AminoAcid.isFullyDefined(q)) {
//	            subs++;
//	        }
//	    }
//	    return subs;
//	}
	
	public static IntList alignSparse(byte[] query, byte[] ref, int maxSubs, IntList seedHits) {
	    if(seedHits==null || seedHits.isEmpty()) {
//	        return align(query, ref, maxSubs); // Fall back to brute force
	    	return null;
	    }
	    
	    IntList results=null;
	    
	    for(int i=0; i<seedHits.size; i++) {
	        int rStart=seedHits.array[i];
	        int subs;
	        
	        // Choose appropriate alignment method based on position
	        if(rStart<0) {
	            subs=alignClipped(query, ref, maxSubs, rStart);
	        } else if(rStart>ref.length-query.length) {
	            subs=alignClipped(query, ref, maxSubs, rStart);
	        } else {
	            subs=align(query, ref, maxSubs, rStart);
	        }
	        
	        if(subs<=maxSubs) {
	            if(results==null) {results=new IntList(4);}
	            results.add(rStart);
	        }
	    }
	    
	    return results;
	}
	
	//This more complex function is ~10% faster
	public static IntList align(byte[] query, byte[] ref, int maxSubs) {
		if(Shared.SIMD && (query.length<256 || maxSubs<256)) {
			return SIMDAlignByte.alignDiagonal(query, ref, maxSubs);
		}
		IntList list=null;
		int rStart=-maxSubs;
		for(; rStart<0; rStart++) {
			int subs=alignClipped(query, ref, maxSubs, rStart);
			if(subs<=maxSubs) {//Rare case
//				(list=(list!=null ? list : new IntList(4))).add(rStart);//1-line version
				if(list==null) {list=new IntList(4);}
				list.add(rStart);
			}
		}
		for(final int limit=ref.length-query.length; rStart<=limit; rStart++) {
			int subs=align(query, ref, maxSubs, rStart);
			if(subs<=maxSubs) {//Rare case
				if(list==null) {list=new IntList(4);}
				list.add(rStart);
			}
		}
		for(final int limit=ref.length-query.length+maxSubs; rStart<=limit; rStart++) {
			int subs=alignClipped(query, ref, maxSubs, rStart);
			if(subs<=maxSubs) {//Rare case
				if(list==null) {list=new IntList(4);}
				list.add(rStart);
			}
		}
		return list;
	}
	
	private static int align(byte[] query, byte[] ref, final int maxSubs, final int rStart) {
		int subs=0;
		for(int i=0, j=rStart; i<query.length && subs<=maxSubs; i++, j++) {
			final byte q=query[i], r=ref[j];
			final int incr=(q!=r || AminoAcid.baseToNumber[q]<0 ? 1 : 0);
			subs+=incr;
		}
		return subs;
	}
	
	private static int alignClipped(byte[] query, byte[] ref, int maxSubs, final int rStart) {
		int subs=0;
		int i=0, j=rStart;
		if(rStart<0) {i-=rStart; j=0; subs-=rStart;}
		for(final int limit=ref.length-query.length; 
				i<query.length && j<=limit && subs<=maxSubs; i++, j++) {
			final byte q=query[i], r=ref[j];
			final int incr=(q!=r || AminoAcid.baseToNumber[q]<0 ? 1 : 0);
			subs+=incr;
		}
		subs+=(query.length-i);
		return subs;
	}
	
	private static int alignClippedSimple(byte[] query, byte[] ref, int maxSubs, final int rStart) {
		int subs=0;
		for(int i=0, j=rStart; i<query.length && subs<=maxSubs; i++, j++) {
			final byte q=query[i], r=(j<0 || j>=ref.length ? (byte)'$' : ref[j]);
			final int incr=(q!=r || AminoAcid.baseToNumber[q]<0 ? 1 : 0);
			subs+=incr;
		}
		return subs;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------         Inner Classes        ----------------*/
	/*--------------------------------------------------------------*/
	
	/** This class is static to prevent accidental writing to shared variables.
	 * It is safe to remove the static modifier. */
	class ProcessThread extends Thread {
		
		//Constructor
		ProcessThread(final ConcurrentReadInputStream cris_, final ByteStreamWriter bsw_, 
				ArrayList<Query> qList, final int maxSubs_, final int tid_){
			cris=cris_;
			bsw=bsw_;
			queries=qList;
			maxSubs=maxSubs_;
			tid=tid_;
		}
		
		//Called by start()
		@Override
		public void run(){
			//Do anything necessary prior to processing
			synchronized(this) {
				//Process the reads
				processInner();

				//Do anything necessary after processing

				//Indicate successful exit status
				success=true;
			}
		}
		
		/** Iterate through the reads */
		void processInner(){
			
			//Grab the first ListNum of reads
			ListNum<Read> ln=cris.nextList();

			//As long as there is a nonempty read list...
			while(ln!=null && ln.size()>0){
//				if(verbose){outstream.println("Fetched "+reads.size()+" reads.");} //Disabled due to non-static access
				
				processList(ln);
				
				//Notify the input stream that the list was used
				cris.returnList(ln);
//				if(verbose){outstream.println("Returned a list.");} //Disabled due to non-static access
				
				//Fetch a new list
				ln=cris.nextList();
			}

			//Notify the input stream that the final list was used
			if(ln!=null){
				cris.returnList(ln.id, ln.list==null || ln.list.isEmpty());
			}
		}
		
		void processList(ListNum<Read> ln){

			//Grab the actual read list from the ListNum
			final ArrayList<Read> refList=ln.list;
			
			//Loop through each read in the list
			for(int idx=0; idx<refList.size(); idx++){
				final Read ref=refList.get(idx);
				
				//Validate reads in worker threads
				if(!ref.validated()){ref.validate(true);}

				//Track the initial length for statistics
				final int initialLength1=ref.length();
				final int initialLength2=ref.mateLength();

				//Increment counters
				readsProcessedT+=ref.pairCount();
				basesProcessedT+=initialLength1+initialLength2;
				
				{
					//Reads are processed in this block.
					processRefSequence(ref);
				}
			}
		}
		
		// Add this method to ProcessThread class
		private IntListHashMap buildReferenceIndex(byte[] ref) {
		    if(!indexQueries || k<=0) {return null;}
		    
		    IntListHashMap index=new IntListHashMap(ref.length*2);
		    
		    final int shift=2*k;
		    final int shift2=shift-2;
		    final int mask=~((-1)<<shift);
		    int kmer=0, rkmer=0;
		    int len=0;
		    
		    for(int i=0; i<ref.length; i++) {
		        final byte b=ref[i];
		        final int x=AminoAcid.baseToNumber[b];
		        final int x2=AminoAcid.baseToComplementNumber[b];
		        kmer=((kmer<<2)|x)&mask;
		        rkmer=((rkmer>>>2)|(x2<<shift2))&mask;
		        
		        if(x<0) {len=0; rkmer=0;} else {len++;}
		        if(len>=k) {
		            // Apply wildcard mask and store both orientations
		            int maskedKmer=(kmer&Query.midMask);
		            int maskedRkmer=(rkmer&Query.midMask);
		            index.put(maskedKmer, i-k+1);
		            if(maskedKmer!=maskedRkmer) {
		                index.put(maskedRkmer, i-k+1);
		            }
		        }
		    }
		    return index;
		}
		
		private IntList getSeedHits(Query q, IntListHashMap refIndex, 
				boolean reverseStrand, IntHashMap hitCounts, String rname) {
//			IntList a=getSeedHitsMap(q, refIndex, reverseStrand, hitCounts);
//			IntList b=getSeedHitsList(q, refIndex, reverseStrand);
//			if(a==null) {assert(b==null) : a+", "+b+", "+new String(q.bases)+", "+rname;}
//			else{assert(a.size==b.size) : a.size+", "+b.size+", "+new String(q.bases);}
//			return b;
			if(useSeedMap) {return getSeedHitsMap(q, refIndex, reverseStrand, hitCounts);}
			else {return getSeedHitsList(q, refIndex, reverseStrand);}
		}
		
		private IntList getSeedHitsList(Query q, IntListHashMap refIndex, boolean reverseStrand) {

			int[] queryKmers=reverseStrand ? q.rkmers : q.kmers;
			if(queryKmers==null) {return null;}
			final int minHits=Math.max(minSeedHits, q.minHits);
			IntList seedHits=new IntList();

			for(int i=0; i<queryKmers.length; i+=qStep) {
				if(queryKmers[i]==-1) {continue;} // Invalid k-mer

				IntList positions=refIndex.get(queryKmers[i]);
				if(positions!=null) {
					for(int j=0; j<positions.size; j++) {
						int refPos=positions.array[j];
						int alignStart=refPos-i; // Adjust for k-mer position in query
//						seedHits.addIfNotEqualToLast(alignStart);//Prevents some duplication
						seedHits.add(alignStart);
					}
				}
			}
			
//			if(verbose) {System.err.println("List: "+seedHits+", "+minHits);}
			
			seedHitsT+=seedHits.size();
			if(seedHits.size<minHits) {return null;}
			// Remove duplicates and sort
			if(seedHits.size>1 || minHits>1) {
				seedHits.sort();
				seedHits.condenseMinCopies(minHits);
			}
			alignmentsT+=seedHits.size();

//			assert(false) : seedHitsT+", "+alignmentsT+", "+minHits;
			return seedHits.isEmpty() ? null : seedHits;
		}
		
		/**
		 * Generates a list of alignment start positions with sufficient seed hit counts.
		 * 
		 * @param q Query sequence to analyze
		 * @param refIndex Reference index for k-mer lookup
		 * @param reverseStrand Flag indicating reverse complement strand
		 * @return List of alignment start positions meeting hit threshold, or null if empty
		 */
		private IntList getSeedHitsMap(Query q, IntListHashMap refIndex, 
				boolean reverseStrand, IntHashMap hitCounts) {
			// Select appropriate k-mer array based on strand
			final int[] queryKmers=reverseStrand ? q.rkmers : q.kmers;
			if(queryKmers==null){return null;}
			final int minHits=Math.max(minSeedHits, q.minHits);

			// Pre-allocate list to store alignment starts meeting hit threshold
			IntList seedHits=null;

			// Iterate through query k-mers at specified step interval
			for(int i=0; i<queryKmers.length; i+=qStep) {
				// Skip invalid k-mers
				if(queryKmers[i]==-1){continue;}

				// Lookup k-mer positions in reference index
				IntList positions=refIndex.get(queryKmers[i]);
				if(positions!=null) {
					seedHitsT+=positions.size;
					if(seedHits==null) {//Lazy allocation
						seedHits=new IntList();
						if(hitCounts==null) {hitCounts=new IntHashMap();}
						else {hitCounts.clear();}
					}
					for(int j=0; j<positions.size; j++) {
						// Calculate potential alignment start position
						int alignStart=positions.array[j]-i;

						// Increment hit count and check if it meets threshold
						int newCount=hitCounts.increment(alignStart);
						if(newCount==minHits) {seedHits.add(alignStart);}
					}
				}
			}
//			if(verbose) {System.err.println("Map: "+hitCounts+", "+minHits);}

			// Return list of seed hits, or null if empty
			alignmentsT+=(seedHits==null ? 0 : seedHits.size());
//			assert(false) : seedHitsT+", "+alignmentsT+", "+minHits;
			return seedHits==null || seedHits.isEmpty() ? null : seedHits;
		}
		
		long processRefSequence(final Read ref) {
			return indexQueries ? processRefSequenceIndexed(ref) : processRefSequenceBrute(ref);
		}

		// Modified processRefSequence method
		long processRefSequenceIndexed(final Read ref){
		    IntListHashMap refIndex=buildReferenceIndex(ref.bases);
		    IntHashMap seedMap=(useSeedMap ? new IntHashMap() : null);
		    
		    long sum=0;
		    for(Query q : queries) {
		        int count=0;

		        // Forward strand alignment
		        IntList seedHits=getSeedHits(q, refIndex, false, seedMap, ref.name());
		        IntList hits=alignSparse(q.bases, ref.bases, maxSubs, seedHits);
		        count+=processHits(q, ref, hits, false, 0);

		        // Reverse strand alignment
		        seedHits=getSeedHits(q, refIndex, true, seedMap, ref.name());
		        hits=alignSparse(q.rbases, ref.bases, maxSubs, seedHits);
		        count+=processHits(q, ref, hits, true, count);
		        
		        readsOutT+=count;
		        basesOutT+=count*q.bases.length;
		        sum+=count;
		    }
		    return sum;
		}
		
		/**
		 * Process a read or a read pair.
		 * @param ref Reference sequence
		 * @return True if the reads should be kept, false if they should be discarded.
		 */
		long processRefSequenceBrute(final Read ref){
			long sum=0;
			for(Query q : queries) {
				int count=0;
				
				// Forward strand alignment
				IntList hits=align(q.bases, ref.bases, maxSubs);
				count+=processHits(q, ref, hits, false, 0);

				// Reverse strand alignment using pre-computed reverse complement query
				hits=align(q.rbases, ref.bases, maxSubs);
				count+=processHits(q, ref, hits, true, count);
				
				readsOutT+=count;
				basesOutT+=count*q.bases.length;
				sum+=count;
			}
			alignmentsT+=(queries.size()*(long)ref.length());
			return sum;
		}

		int processHits(Query q, Read ref, IntList hits, boolean reverseStrand, int count) {
			if(hits==null || hits.size()==0) {return 0;}
			ByteBuilder bb=new ByteBuilder();
			ByteBuilder match=new ByteBuilder(q.bases.length);

			for(int i=0; i<hits.size(); i++) {
				count++;
				int start=hits.get(i);

				// Use appropriate query sequence for match calculation
				byte[] querySeq = reverseStrand ? q.rbases : q.bases;
				toMatch(querySeq, ref.bases, start, match.clear());

				SamLine sl=new SamLine();
				sl.pos=Math.max(start+1, 1);
				sl.qname=q.name;
				sl.setRname(ref.id);
				sl.seq=q.bases; // Always report original query sequence
				sl.qual=q.quals;
				sl.setPrimary(count==1); // Primary if first hit overall
				sl.setMapped(true);
				if(reverseStrand) {sl.setStrand(Shared.MINUS);}
				sl.tlen=q.bases.length;
				sl.cigar=SamLine.toCigar14(match.toBytes(), start, start+q.bases.length-1, ref.length(), q.bases);
				int subs=sl.countSubs();
				sl.addOptionalTag("NM:i:"+subs);
				sl.mapq=Tools.mid(0, (int)(40*(sl.length()*0.5-subs)/(sl.length()*0.5)), 40);
				sl.toBytes(bb).nl();
				if(bb.length()>=16384) {
					if(bsw!=null) {bsw.addJob(bb);}
					bb=new ByteBuilder();
				}
			}
			if(bsw!=null && !bb.isEmpty()) {bsw.addJob(bb);}
			return hits.size;
		}

		void toMatch(byte[] query, byte[] ref, int rStart, ByteBuilder match) {
			for(int i=0, j=rStart; i<query.length; i++, j++) {
				boolean inbounds=(j>=0 && j<ref.length);
				byte q=query[i];
				byte r=(inbounds ? ref[j] : (byte)'$');
				boolean good=(q==r && AminoAcid.isFullyDefined(q));
				match.append(good ? 'm' : inbounds ? 'S' : 'C');
			}
		}

		/** Number of reads processed by this thread */
		protected long readsProcessedT=0;
		/** Number of bases processed by this thread */
		protected long basesProcessedT=0;
		/** Number of alignments done by this thread */
		protected long alignmentsT=0;
		/** Number of seed hits found by this thread */
		protected long seedHitsT=0;
		
		/** Number of reads retained by this thread */
		protected long readsOutT=0;
		/** Number of bases retained by this thread */
		protected long basesOutT=0;
		
		/** True only if this thread has completed successfully */
		boolean success=false;
		
		/** Shared input stream */
		private final ConcurrentReadInputStream cris;
		/** Shared output stream */
		private final ByteStreamWriter bsw;
		private final ArrayList<Query> queries;
		final int maxSubs;
		/** Thread ID */
		final int tid;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------            Fields            ----------------*/
	/*--------------------------------------------------------------*/

	/** Primary input file path */
	private String in1=null;
	/** Secondary input file path */
	private String in2=null;

	/** Primary output file path */
	private String out1=null;
	
	/** Override input file extension */
	private String extin=null;
	/** Override output file extension */
	private String extout=null;
	
	String refFile=null;
	int maxSubs=5;
	int k=13;
	int midMaskLen=1;
	boolean indexQueries=false;
	int qStep=1;
	
	int minSeedHits=1;
	private float minHitsProb=0.9999f;
	boolean useSeedMap=true;
	
	/*--------------------------------------------------------------*/

	/** Number of reads processed */
	protected long readsProcessed=0;
	/** Number of bases processed */
	protected long basesProcessed=0;
	
	protected long alignmentCount=0;
	protected long seedHitCount=0;

	/** Number of reads retained */
	protected long readsOut=0;
	/** Number of bases retained */
	protected long basesOut=0;

	/** Quit after processing this many input reads; -1 means no limit */
	private long maxReads=-1;
	
	/*--------------------------------------------------------------*/
	/*----------------         Final Fields         ----------------*/
	/*--------------------------------------------------------------*/

	/** Primary input file */
	private final FileFormat ffin1;
	/** Secondary input file */
	private final FileFormat ffin2;
	
	/** Primary output file */
	private final FileFormat ffout1;
	
	@Override
	public final ReadWriteLock rwlock() {return rwlock;}
	private final ReadWriteLock rwlock=new ReentrantReadWriteLock();
	
	/*--------------------------------------------------------------*/
	/*----------------           Statics            ----------------*/
	/*--------------------------------------------------------------*/
	
	/*--------------------------------------------------------------*/
	/*----------------        Common Fields         ----------------*/
	/*--------------------------------------------------------------*/
	
	/** Print status messages to this output stream */
	private PrintStream outstream=System.err;
	/** Print verbose messages */
	public static boolean verbose=false;
	/** True if an error was encountered */
	public boolean errorState=false;
	/** Overwrite existing output files */
	private boolean overwrite=true;
	/** Append to existing output files */
	private boolean append=false;
	
}
